# Lips.py  # 2022-11-22
# MIT LICENSE 2020 Ewerton R. Vieira

import scipy.optimize as optimize
import numpy as np
from datetime import datetime


# TODO imcorporate
"""def sampled_Lipschitz(lower_bounds, upper_bounds, N, f, base_name=""):
    x = sample_points(lower_bounds, upper_bounds, N)
    K = 0
    for i in range(len(x)):
        x0 = x[i]
        for j in range(i + 1, len(x)):
            x1 = x[j]
            y0 = np.array(f(x0))
            y1 = np.array(f(x1))
            K_temp = np.linalg.norm(y1 - y0)/np.linalg.norm(x1 - x0)
            if K_temp > K:
                K = K_temp

    dir_path = os.path.abspath(os.getcwd()) + "/output/"
    Lipschitz_file = dir_path + base_name + "_sampled_Lipschitz.txt"
    np.savetxt(Lipschitz_file, [K])

    return [K for a in range(len(lower_bounds))]"""


class Lips:

    def dy_dx(self, x0, x1):
        y0 = np.array(self.g(x0))
        y1 = np.array(self.g(x1))
        x0 = np.array(x0)
        x1 = np.array(x1)
        return np.linalg.norm(y1 - y0), np.linalg.norm(x1 - x0)

    def slope(self, func, x0, x1):
        y0 = np.array(func(x0))
        y1 = np.array(func(x1))
        x0 = np.array(x0)
        x1 = np.array(x1)
        return - np.linalg.norm(y1 - y0)/np.linalg.norm(x1 - x0)

    def f(self, x):
        size = len(x)
        x0 = x[0: size//2]
        x1 = x[size//2: size]
        return self.slope(self.g, x0, x1)

    def Powell(self, x_initial):

        res = optimize.minimize(self.f, x_initial, method='Powell',
                                bounds=self.bnds, options={'xtol': self.error, 'ftol': self.error, 'maxiter': self.N, 'maxfev': self.N})
        # print(res)
        return [res.x, res.fun]

    def acceptance_probability(self, E, E_new, T):
        # if E_new < E:
        #     return 1
        return np.exp(-(E_new-E)/T)

    def random_neighbour(self, x):
        for i in range(len(x)):
            x[i] += np.random.uniform(-1, 1)
            if x[i] < self.bnds[i][0]:
                x[i] = self.bnds[i][0]
            elif x[i] > self.bnds[i][1]:
                x[i] = self.bnds[i][1]
        return x

    def random_x(self):
        return [(self.bnds[i][1] - self.bnds[i][0])*np.random.random() + self.bnds[i][0] for i in range(len(self.bnds))]

    def simulated_annealing(self, steps):
        x = self.random_x()
        X, E = self.Powell(x)
        # print("x=%.2f, fmin=%.2f" % (x, E))
        for k in range(steps):
            T = (1 - k/steps)
            x = self.random_neighbour(x)
            E_new = self.Powell(x)[1]
            X_new = self.Powell(x)[0]
            P = self.acceptance_probability(E, E_new, T)
            if P > np.random.random():
                E = E_new
                X = X_new
            # print("x=%.4f, fmin=%.4f, Prob.=%.4f" % (x, E, P))
        return abs(E), X

    def find_and_save_Lipschitz(self, n):

        startTime = datetime.now()
        E, X = self.simulated_annealing(n)
        print(f"Time to find Lipschitz constant = {datetime.now() - startTime}")

        x_0 = [X[a] for a in range(len(self.bnds)//2)]
        x_1 = [X[a] for a in range(len(self.bnds)//2, len(self.bnds))]
        # print(x_0)
        # print(x_1)
        info = f"Lipschit constant={E}, \nx_0={x_0}, x_1={x_1} \nf(x_0)={self.g(x_0)}, f(x_1)={self.g(x_1)}, \ndy and dx={self.dy_dx(x_0, x_1)} \n "
        print(info)
        file_name = "Lip" + self.base_name + ".txt"
        with open(file_name, "a") as file:
            file.write(info)
        return E

    def __init__(self, lower_bounds, upper_bounds, g, base_name, error=0.0001, N=1):

        self.base_name = base_name
        self.g = g
        self.bnds = [(lower_bounds[i], upper_bounds[i]) for i in range(len(lower_bounds))]
        self.bnds = self.bnds + self.bnds
        self.error = error
        self.N = N * len(self.bnds) * 1000  # default for Powell's mimimize

        # error=0.000001
        # N=10


# finding_Lipschitz(lower_bounds, upper_bounds, g, base_name)

# print(slope(g, [-2.34279779, - 2.51941202, - 2.56941876],
# [-2.44407475, - 2.27628564, - 2.55924496]))

# print(g([-2.34279779, - 2.51941202, - 2.56941876]))

# print(g([-2.44407475, - 2.27628564, - 2.55924496]))


# [2.55040391 2.36068774 0.28470773 2.75396691 2.36613446 0.29615437]
# -36.94207095524288


# def F(rect):
#     return F_K(rect, g, K)
# run_CMGDB(subdiv_min, subdiv_max, lower_bounds, upper_bounds, phase_periodic, F, base_name)



# MultivaluedMap.py  # 2021-20-10
# MIT LICENSE 2020 Ewerton R. Vieira
import pychomp
import graphviz
import dytop.Poset as Poset


def Dict_inverse(D):
    """Return the inverse of a dictionary D"""
    inv_map = {}
    for k, v in D.items():
        for i in v:
            inv_map[i] = inv_map.get(i, []) + [k]
    return inv_map


class MultivaluedMap:

    def __init__(self, map_graph, morse_graph):
        """
        Adding features to map_graph to become a mvm_graph
        """

        # self.dir_path = os.path.abspath(os.getcwd()) + "/output/"

        self.morse_graph = morse_graph

        # Get number of vertices
        num_verts = map_graph.num_vertices()

        self.vertices_ = {a for a in range(num_verts)}

        self.map_graph = map_graph

        # self.mapping: map_graph -> condensation_graph
        condensation_graph, self.mapping = pychomp.CondensationGraph(
            self.vertices_, map_graph.adjacencies)

        self.CG = condensation_graph
        # self.CG.graphviz = self.condensation_graph_with_labels

        self.mapping_inv = {}  # inverse of mapping condensation_graph -> map_graph
        for key, value in self.mapping.items():
            self.mapping_inv.setdefault(value, []).append(key)

        # condensation_graph with poset structure
        self.condensation_graph = Poset.Poset(condensation_graph, save_memory=True)

        MG = pychomp.DirectedAcyclicGraph()  # building Morse Graph Poset
        MG.add_vertex(0)
        for u in range(morse_graph.num_vertices()):
            for v in morse_graph.adjacencies(u):
                MG.add_edge(u, v)
        self.MG = Poset.Poset(MG)

    def vertices(self):
        """
        Return the set of elements in the poset
        """
        return self.vertices_

    def upset(self, v):
        """
        Return the set { u : u >= v }
        """
        U = []
        for a in self.condensation_graph.ancestors(self.mapping[v]).union({self.mapping[v]}):
            U = U + self.mapping_inv[a]
        return set(U)

    def mvm_graph(self):
        mvm = pychomp.DirectedAcyclicGraph()  # building domain graph
        mvm.add_vertex(0)
        for v in self.vertices():
            for u in self.map_graph.adjacencies(v):
                mvm.add_edge(v, u)
        return mvm

    def build_viz(self, graph, shape='circle'):
        """ Return a graphviz string describing the graph and its labels """
        gv = 'digraph {\n'

        gv += f'node [fontsize=12, shape={shape}]'

        for v in graph.vertices():
            gv += f"{v}[label={v}];\n"

        for v in graph.vertices():
            for u in graph.adjacencies(v):
                gv += f"{v}->{u};\n"
        return gv + '}\n'

    def condensation_graph_with_labels(self):
        """ Add labels to the condensation_graph"""
        gv = 'digraph {\n'

        gv += f'node [fontsize=12, shape=square]\n'

        for v in self.CG.vertices():
            inv_v = self.mapping_inv[v]
            if len(inv_v) == 1 and inv_v[0] in self.mvm_graph().adjacencies(inv_v[0]):
                gv += f'{v}[label=\"<{inv_v[0]}>\"];\n'
            else:
                temp = ''.join(f'{e} ' for e in self.mapping_inv[v])
                gv += f'{v}[label=\"{temp}\"];\n'

        for v in self.CG.vertices():
            for u in self.CG.adjacencies(v):
                gv += f"{v}->{u};\n"
        gv += "}\n"

        return gv
        # return graphviz.Source(gv)

    def _repr_svg_(self):
        """
        Return svg representation for visual display
        """
        # return graphviz.Source(self.mvm_graph().graphviz())._repr_svg_()

        # f = graphviz.Source(self.graphviz(), filename=self.filename, format='pdf')
        # f.render(cleanup=True);
        # return graphviz.Source(str(self.build_viz(self.mvm_graph())))
        return graphviz.Source(self.build_viz(self.mvm_graph()))
    
    def vizualize_CG(self):
        return graphviz.Source(self.condensation_graph_with_labels())

    def vizualize(self):
        return graphviz.Source(self.build_viz(self.mvm_graph()))



# PlotRoA.py  # 2022-11-22
# MIT LICENSE 2020 Ewerton R. Vieira

import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
import matplotlib
import os
import csv
import numpy as np


def PlotRoA(lower_bounds, upper_bounds, selection=[], fig_w=8, fig_h=8, xlim=None, ylim=None, fontsize=16,
              cmap=matplotlib.cm.get_cmap('viridis', 256), name_plot="", from_file=None, plot_point=False, section=None, from_file_basic=False, dir_path=""):
    """ TODO:
    * section = ([z,w],(a,b,c,d)), 3D section when [z,w]=(c,d)
    * selection = selection of morse sets
    * check 1D and 3D plottings
    * check save file"""

    dim = len(lower_bounds)

    eps_round_up = 0 # 0.000000005 parameter that remove white lines from the plot for some cases.

    # path to save and read files
    if not dir_path:
        dir_path = os.path.abspath(os.getcwd()) + "/output/"

    # read file saved by RoA
    if from_file and not from_file_basic:

        # read file
        from_file = dir_path + from_file + "_RoA_" + ".csv"
        with open(from_file, "r") as file:
            f = csv.reader(file, delimiter=',')
            next(f)
            box_size = [float(i) for i in next(f)[0:dim]]
            next(f)
            Tiles = []
            Morse_nodes = []
            Boxes = []
            num_morse_sets = 0
            counter_temp = 0
            for row in f:

                if row[0] == "Tile_in_Morse_set":
                    counter4morse_sets = counter_temp
                    continue
                counter_temp += 1
                Tiles.append(int(row[0]))
                Morse_nodes.append(int(row[1]))
                Boxes.append([float(a) for a in row[2:2+2*dim]])

                if num_morse_sets < int(row[1]):   # find the num_morse_sets - 1
                    num_morse_sets = int(row[1])

            num_morse_sets += 1

        # print(Tiles, Morse_nodes, Boxes)
        variables = [a for a in range(dim)]

        if not selection:
            selection = [i for i in range(num_morse_sets)]

        cmap_norm = matplotlib.colors.Normalize(vmin=0, vmax=num_morse_sets-1)

        morse = {}  # tiles in morse sets
        tiles = {}  # tiles in regions of attraction (not including tiles in morse set)

        volume_cube = 1
        d_vol = dict()
        for i in range(dim):
            volume_cube *= box_size[i]

        for i, m_node in enumerate(Morse_nodes):
            if m_node not in selection:  # only add the selected Morse sets
                continue
            clr = matplotlib.colors.to_hex(cmap(cmap_norm(m_node)))

            if i < counter4morse_sets:  # associate center of boxes to the Morse tiles
                B = tiles.get(clr, [])
                B.append(Boxes[i])
                tiles[clr] = B
                # compute the total volume of Morse tiles
                d_vol[m_node] = d_vol.get(m_node, 0) + volume_cube

            else:  # associate  boxes to the Morse sets
                A = morse.get(clr, [])
                A.append(Boxes[i])
                morse[clr] = A
                d_vol[m_node] = d_vol.get(m_node, 0) + volume_cube

            # # compute the total volume of Morse tiles
            # d_vol[m_node] = d_vol.get(m_node, 0) + volume_cube

        print(f'dictionary with volume of all Morse tiles = {d_vol}')
    # read file saved by RoA

    # read file saved by CMGDB (only Morse tiles)
    if from_file and from_file_basic:

        from_file = dir_path + from_file + ".csv"
        morse = {}
        with open(from_file, "r") as file:
            f = csv.reader(file, delimiter=',')
            Morse_nodes = []
            box = []
            for row in f:
                dim = len(row)//2
                Morse_nodes.append(int(float(row[-1])))
                box.append([float(row[i]) for i in range(2*dim)])
            box_size = [float(row[i+dim]) - float(row[i]) + eps_round_up for i in range(dim)]

        variables = [a for a in range(dim)]
        num_morse_sets = Morse_nodes[-1] + 1

        cmap_norm = matplotlib.colors.Normalize(vmin=0, vmax=num_morse_sets-1)

        morse = {}
        tiles = {}
        for i, m_node in enumerate(Morse_nodes):
            clr = matplotlib.colors.to_hex(cmap(cmap_norm(m_node)))
            A = morse.get(clr, [])
            A.append(box[i])
            morse[clr] = A

        tiles = morse
    # read file saved by CMGDB (only Morse tiles)

    # for dim 1, add fake dimension
    if dim == 1:
        size1d = 64
        box_size.append(fig_h/size1d)
        # dim = 2
        lower_bounds += lower_bounds
        upper_bounds += upper_bounds
        variables = [0,1]

    # 2D plotting or 2D with a given section
    if section or dim <= 2:
        fig, ax = plt.subplots(figsize=(fig_w, fig_h))

        if dim <= 2:
            section = ([], 'projection')

        variables_section = list(set(variables) - set(section[0]))
        d1 = variables_section[0]
        d2 = variables_section[1]

        if section[1] == 'projection':
            section = ([], 'projection')  # clean section to do projection

        if not from_file_basic:
            for i, j in tiles.items():
                rectangles_list = []  # set instead of list for avoiding repetition
                for row in j:
                    if section[1] == 'projection':
                        in_section = [True]

                    else:
                        in_section = [row[k] - box_size[k]/2 <= section[1][k]
                                      < row[k] + box_size[k]/2 for k in section[0]]

                    if all(in_section):
                        ### Recompute box_size to work for adaptive scheme
                        box_size = [float(row[i+dim]) - float(row[i]) + eps_round_up for i in range(dim)]
                        if dim == 1:
                            box_size.append(fig_h/size1d)
                            row[d2] = -fig_h/(2*size1d)
                        ###


                        rectangle = Rectangle((row[d1], row[d2]), box_size[d1], box_size[d2])
                        rectangles_list.append(rectangle)
                pc = PatchCollection(rectangles_list, cmap=cmap, fc=i, alpha=0.4, ec='none')
                ax.add_collection(pc)

        for i, j in morse.items():
            rectangles_list = []
            for row in j:

                if section[1] == 'projection':
                    in_section = [True]

                else:
                    in_section = [row[k] - box_size[k]/2 <= section[1][k]
                                  < row[k] + box_size[k]/2 for k in section[0]]

                if all(in_section):
                    ### Recompute box_size to work for adaptive scheme
                    box_size = [float(row[i+dim]) - float(row[i]) + eps_round_up for i in range(dim)]
                    if dim == 1:
                        box_size.append(fig_h/size1d)
                        row[d2] = -fig_h/(2*size1d)
                    ###
                    rectangle = Rectangle((row[d1], row[d2]), box_size[d1], box_size[d2])
                    rectangles_list.append(rectangle)
            pc = PatchCollection(rectangles_list, cmap=cmap, fc=i, alpha=1, ec='none')
            ax.add_collection(pc)

        tick = 5  # tick for 2D plots

        if xlim and ylim:
            ax.set_xlim([xlim[0], xlim[1]])
            ax.set_ylim([ylim[0], ylim[1]])
            plt.xticks(np.arange(xlim[0], xlim[1], tick))
            plt.yticks(np.arange(ylim[0], ylim[1], tick))
            plt.xticks(np.linspace(xlim[0], xlim[1], tick))
            plt.yticks(np.linspace(ylim[0], ylim[1], tick))
        else:
            ax.set_xlim([lower_bounds[d1], upper_bounds[d1]])
            ax.set_ylim([lower_bounds[d2], upper_bounds[d2]])
            plt.xticks(np.arange(lower_bounds[d1], upper_bounds[d1], tick))
            plt.yticks(np.arange(lower_bounds[d2], upper_bounds[d2], tick))
            plt.xticks(np.linspace(lower_bounds[d1], upper_bounds[d1], tick))
            plt.yticks(np.linspace(lower_bounds[d2], upper_bounds[d2], tick))
        plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)
        ax.xaxis.label.set_size(fontsize)
        ax.yaxis.label.set_size(fontsize)

        ax.set_xlabel(str(d1))
        ax.set_ylabel(str(d2))
        if section[1] == 'projection':
            value_section = tuple([0 for i in section[0]])

        else:
            value_section = tuple([int(section[1][i]*100) for i in section[0]])
        name_plot = f'{dir_path}{name_plot}'
        name_plot += f'_{section[0]}_{value_section}' if section[0] else ''
    # 2D plotting or 2D with a given section

    # 3D plotting
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        # fig, ax = plt.subplots(figsize=(fig_w, fig_h))
        #

        if plot_point:
            for i, j in morse.items():
                j = np.array(j)
                plt.plot(j[:, 0], j[:, 1], j[:, 2], 's', color=i, alpha=0.1)

            for i, j in tiles.items():
                j = np.array(j)
                plt.plot(j[:, 0], j[:, 1], j[:, 2], 's', color=i, alpha=1)

        else:

            voxel_grid_x = int(np.rint((upper_bounds[0]-lower_bounds[0])/box_size[0]))
            voxel_grid_y = int(np.rint((upper_bounds[1]-lower_bounds[1])/box_size[1]))
            voxel_grid_z = int(np.rint((upper_bounds[2]-lower_bounds[2])/box_size[2]))

            voxel_grid = np.zeros((voxel_grid_x, voxel_grid_y, voxel_grid_z))

            x, y, z = np.indices(np.array(voxel_grid.shape)+1)

            x = box_size[0]*x + lower_bounds[0]
            y = box_size[1]*y + lower_bounds[1]
            z = box_size[2]*z + lower_bounds[2]

            # for i, j in tiles.items():
            #     for row in j:
            #         v_x = int(np.rint((row[0] - box_size[0]/2 - lower_bounds[0]) / box_size[0]))
            #         v_y = int(np.rint((row[1] - box_size[1]/2 - lower_bounds[1]) / box_size[1]))
            #         v_z = int(np.rint((row[2] - box_size[2]/2 - lower_bounds[2]) / box_size[2]))
            #         voxel_grid[v_x, v_y, v_z] = True
            #     # ax.voxels(x, y, z, voxel_grid, facecolors=i, alpha=0.2)
            #     voxel_grid = np.where(voxel_grid == True, False, True)
            #     ax.voxels(x, y, z, voxel_grid, facecolors=i, alpha=0.2)
            #     voxel_grid = np.zeros((voxel_grid_x, voxel_grid_y, voxel_grid_z))

            for i, j in morse.items():
                for row in j:
                    v_x = int(np.rint((row[0] - box_size[0]/2 - lower_bounds[0]) / box_size[0]))
                    v_y = int(np.rint((row[1] - box_size[1]/2 - lower_bounds[1]) / box_size[1]))
                    v_z = int(np.rint((row[2] - box_size[2]/2 - lower_bounds[2]) / box_size[2]))
                    voxel_grid[v_x, v_y, v_z] = True
                # ax.voxels(x, y, z, voxel_grid, facecolors=i, alpha=0.2)
                voxel_grid = np.where(voxel_grid == True, True, True)
                ax.voxels(x, y, z, voxel_grid, facecolors=i, alpha=0.8)
                voxel_grid = np.zeros((voxel_grid_x, voxel_grid_y, voxel_grid_z))

        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Theta')
        name_plot = dir_path + name_plot

    # 3D plotting

    # save file with name_plot
    
    if name_plot != dir_path:
        print("here")
        plt.savefig(name_plot)

    return fig, ax

    # return fig, ax, d_vol


# Bistable_RoA.py  # 2022-20-01
# MIT LICENSE 2020 Ewerton R. Vieira
"""Convert any system with two attractors into a bistable MG"""
import RoA
import matplotlib.pyplot as plt

def assignment(order_retraction):
    assign = dict()
    for k, v in order_retraction.items():
        for i in v:
            assign[i] = k
    assign[-1]=-1
    return assign

def update_file(order_retraction, file_input, name="bistable_RoA.csv"):
    assign = assignment(order_retraction)

    with open(name, "w") as file_out:
        with open(file_input, "r") as file:
            # for i in range(3):
            #     line = file.readline()
            #     file_out.writelines(line)
            line = file.readline()
            while line != "":
                line = line.split(",")
                if line[0][0:4]!="Tile" and len(line)>3:
                    line[1] = str(assign[int(line[1])])
                line = ",".join(line)
                file_out.writelines(line)
                line = file.readline()



order_retraction = {1:{3}, 0:{0,1,2,8,5,6}, 2:{4,7,9,10,11}}
name = "output/bistable_RoA_.csv"

assign = assignment(order_retraction)

update_file(order_retraction, "output/MG_RoA_.csv", name)



lower_bounds = [-1,-1]
upper_bounds = [1,1]
fig, ax = RoA.PlotTiles(lower_bounds, upper_bounds,
            from_file="bistable")

out_pic = "output/bistable_RoA_"

plt.savefig(out_pic, bbox_inches='tight')


# Grid.py  # 2022-11-22
# MIT LICENSE 2020 Ewerton R. Vieira
import numpy as np
import csv

class Grid:
    def __init__(self, lower_bounds, upper_bounds, subdivision, base_name=""):
        self.lower_bounds = lower_bounds
        self.upper_bounds = upper_bounds
        self.subdivision = subdivision
        self.base_name = base_name
        self.dim = len(lower_bounds)
        self.subdiv = self.coordinate_subdiv()

        #size of the box for the uniform initial subdivision
        self.size_of_box = [(upper_bounds[a] - lower_bounds[a])/(2**self.subdiv[a]) for a in range(self.dim)]

    def coordinate_subdiv(self):
        base_sub = self.subdivision // self.dim  # equal subdivision for all variables
        subdiv = [base_sub for i in range(self.dim)]
        remainder = self.subdivision % self.dim
        for i in range(remainder):  # adding extra 1 from each variables when dim doesnt subdivide subdivision
            subdiv[i] += 1
        return subdiv

    def coordinates2index(self, coordinate):
        """Given cell coordinates return the index of a
        cell based on CMGDB. Only works for uniform grid"""
        remainder = self.subdivision % self.dim
        added_remainder = [(remainder + self.dim - 1 - a) % self.dim for a in range(self.dim)]
        sum = 0
        for i in range(self.dim):
            c = coordinate[i]
            count = 0
            while c:
                sum += (c & 1) * 2 ** (added_remainder[i] + count*self.dim)
                c = c >> 1
                count +=1
        return sum

    def point2cell_coord(self, point):
        """Return coordinate of cell on grid that contains the input.
        Caution:coordinate of cell is equal to coordinate of vertex on grid for region_of_cube 0"""

        coordinate = []
        for i in range(self.dim):

            coordinate_temp_i = int((point[i] - self.lower_bounds[i]) / self.size_of_box[i])

            # update coordinate in case it falls on the right fringe (right boundary)
            if coordinate_temp_i == 2**self.subdiv[i]:
                coordinate_temp_i -= 1
            coordinate.append(coordinate_temp_i)

        return coordinate

    def point2cell(self, point):
        """Return the cell id on grid that contains the input point
        Caution:id of cell is equal to id of vertex on grid for region_of_cube 0"""
        coordinate = self.point2cell_coord(point)
        return self.get_id_vertex(coordinate, self.dim, self.subdivision)

    def write_map_grid(self, f, base_name="", write_regions=False):
        """write a csv file with image of all vertex of a grid
        given by lower_bounds, upper_bounds, subdivision
        Input:
        f : funtion
        lower_bounds, upper_bounds, subdivision : information about the cubical grid
        base_name : name of the output file
        write_regions : add lines with regions information (useful to check by hand)"""

        base_name += ".csv"
        with open(base_name, 'w') as file_:
            file = csv.writer(file_)
            file.writerow([self.lower_bounds, self.upper_bounds, self.subdivision])

            for region_index in range(2 ** self.dim):

                if write_regions:
                    region_of_cube = f"{region_index:b}"
                    file.writerow([region_of_cube])

                sub_subdivision, dim_face = self.subdivision_of_face(region_index)

                # print(region_index, sub_subdivision)
                for id in range(2**sub_subdivision):

                    position = self.position_at_grid(id, dim_face, sub_subdivision)
                    x = self.grid_vertex2vertex(region_index, position)

                    # x = position
                    file.writerow(f(x))

    def grid_vertex2vertex(self, face, position):

        count = count_for_face = 0
        x = []
        while count < self.dim:
            if (face & 1):
                x.append(self.upper_bounds[count])
                count_for_face -= 1
            else:
                x.append(self.lower_bounds[count] + self.size_of_box[count]*position[count_for_face])
            face >>= 1
            count += 1
            count_for_face += 1

        return x

    def subdivision_of_face(self, face):
        # subdivision on a face (example face = 0101 has 2**2 subdivisions)
        count = sub_subdivision = dim_face = 0
        face = 2**self.dim + ~face  # flip bits with cap size with 2**dim
        while face:
            if (face & 1):
                sub_subdivision += self.subdiv[count]
                dim_face += 1
            face >>= 1
            count += 1

        return sub_subdivision, dim_face

    def position_at_grid(self, id, dim, subdivision):
        """Given a number (id of a vertice), return the coordinates"""
        position = [0 for i in range(dim)] #[0] * dim
        index = subdivision

        if id >= 2**subdivision:
            return "invalide id for vertice"

        count = 0
        while id or index:
            position[count % dim] <<= 1
            position[count % dim] += (id & 1)
            count += 1
            id >>= 1
            index -= 1
        return position

    def get_id_vertex(self, position, dim, subdivision):
        """Given the position of a vertex, return the id, inverse of position_at_grid"""
        # if len(position) != dim:
        #     return "invalide position or/and dimension"
        position_temp = list(position)
        count = subdivision
        id = 0
        while any(position_temp) or count:
            id <<= 1
            j = (count - 1) % dim
            id += position_temp[j] & 1
            position_temp[j] >>= 1
            count -= 1
        return id

    def vertex2grid_vertex(self, vertex):
        """Return the region_of_cube and the coordinates of a point (geometric vertex) at grid"""

        coordinate = [int(
            np.rint((vertex[a] - self.lower_bounds[a]) / self.size_of_box[a])
        ) for a in range(self.dim)
        ]
        # print(f"coordinate in the cube {coordinate}")

        region_of_cube = [coordinate[a] == 2**self.subdiv[a] for a in range(self.dim)]

        # print(region_of_cube)

        coordinate = [coordinate[a] - region_of_cube[a]
                      for a in range(self.dim) if not region_of_cube[a]]  # update coordinate for the region_of_cube

        if not coordinate:
            coordinate = [0]

        # print(f"coordinate in the region {coordinate}")

        region_of_cube_bit_form = 0  # from list to bit [1,0,1] to 5
        for index, value in enumerate(region_of_cube):
            region_of_cube_bit_form += (value << index)

        # print(region_of_cube, region_of_cube_bit_form, coordinate)

        sub_subdivision, dim_face = self.subdivision_of_face(region_of_cube_bit_form)

        return region_of_cube_bit_form, self.get_id_vertex(coordinate, dim_face, sub_subdivision)

    def load_map_grid(self, file_name):
        map = []
        with open(file_name, 'r') as file_:
            next(file_)
            file = csv.reader(file_, quoting=csv.QUOTE_NONNUMERIC)
            for row in file:
                map.append(row)
        return map

    def image_of_vertex_from_loaded_map(self, map, vertex):
        """input:
        subdivision can be smaller than subdivision used to compute map
        TODO: for now subdivision is the same to compute map (fix this)
        """
        region_of_cube, id = self.vertex2grid_vertex(vertex)

        # print(region_of_cube)

        region_position_file = 0
        for i in range(region_of_cube):
            # if count:
            num = 0
            for index, value in enumerate(self.bit_to_list(i, self.dim)):
                num += self.subdiv[index] * (not value)
                # not value since 0 counts as interior and it should take in account
            region_position_file += 1 << num

        # print("region and id", region_of_cube, id)
        # print("region_position_file + id", region_position_file + id)
        return map[region_position_file + id]

    def list_to_bit(self):
        return None

    def bit_to_list(self, number, dim):
        return [(number >> i) & 1 for i in range(dim)]

    def id2image(self, data):
        """Input: data (x,f(x))
        Return a list indexed by the cell id on CMGDB where each element
        of the returned list correspond to a list of images f(x)
        for all x in cell id \cap data"""
        id2image_ = [list() for i in range(2 ** self.subdivision)]
        data_x = data[:, 0:self.dim].reshape(-1,self.dim)
        data_fx = data[:, self.dim::].reshape(-1,self.dim)

        for i in range(len(data_x)):
            id2image_[self.point2cell(data_x[i])].append(data_fx[i])

        return id2image_


    def valid_grid(self, data, transform = lambda x : x, neighbors=True):
        """initializing valid list of cell on the grid, it is indexed by the
        number assiging to each cell on the grid in CMGDB"""
        valid_list = [False]*(2**self.subdivision)

        if neighbors:
            for i in range(len(data)):
                z = transform(data[i]).tolist()
                all_neighbors_position = self.neighbors(self.point2cell(z))
                for neighbor_position in all_neighbors_position:
                    # valid_list[self.point2cell(neighbor_position)] = True
                    valid_list[self.get_id_vertex(neighbor_position, self.dim, self.subdivision)] = True
        else:
            for i in range(len(data)):
                z = transform(data[i]).tolist()
                valid_list[self.point2cell(z)] = True
        return valid_list

    def uniform_sample_old(self):
        """Return uniform samples on a given grid"""

        number_extra_subdivisions = 0 # add one extra subdivision for each variable
        size_subdivision = self.subdivision + number_extra_subdivisions*self.dim
        size_of_box = [(self.upper_bounds[a] - self.lower_bounds[a])/(2**(self.subdiv[a]+number_extra_subdivisions)) for a in range(self.dim)]

        data = []
        for i in range(2**(size_subdivision)):
            coordinate = self.position_at_grid(i, self.dim, size_subdivision)
            values = [self.lower_bounds[a] + size_of_box[a] * coordinate[a] for a in range(self.dim)]
            data.append(values)

        data = np.array(data)
        return data


    def uniform_sample(self):
        """Return uniform samples on a given grid"""
        data = []
        for region_index in range(2 ** self.dim):

            sub_subdivision, dim_face = self.subdivision_of_face(region_index)
            for id in range(2**sub_subdivision):
                position = self.position_at_grid(id, dim_face, sub_subdivision)
                x = self.grid_vertex2vertex(region_index, position)
                data.append(x)
        return np.array(data)
    
    def point2indexCMGDB(self, point):
        """Given a point return the index (id) in the CMGDB"""
        cell = self.point2cell(point)
        return self.coordinates2index(self.position_at_grid(cell, self.dim, self.subdivision))

    def neighbors(self, cell_index):
        """Given index of region 0, return the neighbors in region 0 (squared ball with radius 1)"""
        coordinate = self.position_at_grid(cell_index, self.dim, self.subdivision) 

        add_elements = [[0] for i in range(self.dim)]

        vector_limit = []

        total_add_elements = 1
        for i in range(self.dim):
            if coordinate[i] == 2**self.subdiv[i] - 1:  # only in region 0
                add_elements[i]+=[-1]
                total_add_elements *= 2
                vector_limit += [2]
            elif coordinate[i] == 0:
                add_elements[i]+=[1]
                total_add_elements *= 2
                vector_limit += [2]
            else:
                add_elements[i]+=[-1,1]
                total_add_elements *= 3
                vector_limit += [3]

        all_neighbors_position = []

        index = 0
        while index < total_add_elements:
            new_coordinate = []
            quocient = index
            for i in range(self.dim):  #index to vector in \oplus_i Z_vector_limit[i]
                reminder = quocient % vector_limit[i]
                quocient = quocient // vector_limit[i]
                new_coordinate += [coordinate[i] + add_elements[i][reminder]]
            all_neighbors_position += [new_coordinate]
            index += 1
        
        return all_neighbors_position


# Morse_graph_updated.py  # 2023-01-22
# MIT LICENSE 2020 Ewerton R. Vieira

class Morse_graph_updated:
    def __init__(self, morse_graph, MG_updated):

        self.morse_graph = morse_graph

        self.MG_ = Poset.Poset(MG_updated)

        self.MG_updated_ = pychomp.DirectedAcyclicGraph()  # building Morse Graph Poset
        self.MG_updated_.add_vertex(0)

        for u in self.morse_graph.vertices():
            for v in self.MG_.parents(u):
                self.MG_updated_.add_edge(v, u)  # inverse since RoA build MG_updated inverted

        # for (u,v) in self.morse_graph.edges():
        #     if (u,v) in MG_updated.edges():
        #         self.MG_updated_.add_edge(u, v)

    def adjacencies(self, i):
        return self.MG_updated_.adjacencies(i)

    def adjacencies_unreduced(self, i):
        return self.MG_updated_.adjacencies_unreduced(i)

    def annotations(self, i):
        return self.morse_graph.annotations(i)

    def edges(self):
        return self.MG_updated_.edges()

    def edges_unreduced(self):
        return self.MG_updated_.edges_unreduced()

    def morse_set(self, i):
        return self.morse_graph.morse_set(i)

    def morse_set_boxes(self, i):
        return self.morse_graph.morse_set_boxes(i)

    def num_vertices(self):
        return self.morse_graph.num_vertices()

    def phase_space_box(self, i):
        return self.morse_graph.phase_space_box(i)

    def vertices(self):
        return self.morse_graph.vertices()


# Poset.py  # 2021-20-10
# Base on Shaun Poset.py
# MIT LICENSE 2020 Ewerton R. Vieira

from pychomp.DirectedAcyclicGraph import *
import graphviz


class Poset:  # Represent a Poset

    def ListOfNivelSets(self):  # return the list of increasing Nivel sets of a given poset

        P0 = {v for v in self.vertices() if len(self.children(v)) == 0}

        def NivelSet(P_i):  # return the parent set of a given set
            P_iplus = set()
            for a in P_i:
                # print(str(a)+' -> '+ str(CMG.parents(a)))
                P_iplus = P_iplus.union(self.parents(a))
            return P_iplus - P_i

        P = [P0]
        j = 0
        while P[j] != set():
            P.append(NivelSet(P[j]))
            j += 1

        return P

    def __init__(self, graph, save_memory=False):
        """
        Create a Poset P from a DAG G such that x <= y in P iff there is a path from x to y in G
        """
        self.vertices_ = set(graph.vertices())

        if save_memory:  # to save memory it doesnt build descendants
            self.ancestors_ = graph.transpose()
            self.ancestors_ = self.ancestors_.transitive_closure()

        else:
            self.descendants_ = graph.transitive_closure()
            self.ancestors_ = self.descendants_.transpose()
            self.children_ = graph.transitive_reduction()
            self.parents_ = self.children_.transpose()

    def __iter__(self):
        """
        Allows for the semantics
          [v for v in poset]
        """
        return iter(self.vertices())

    def vertices(self):
        """
        Return the set of elements in the poset
        """
        return self.vertices_

    def parents(self, v):
        """
        Return the immediate predecessors of v in the poset
        """
        return self.parents_.adjacencies(v)

    def children(self, v):
        """
        Return the immediate successors of v in the poset
        """
        return self.children_.adjacencies(v)

    def ancestors(self, v):
        """
        Return the set { u : u > v }
        """
        return self.ancestors_.adjacencies(v)

    def upset(self, v):
        """
        Return the set { u : u >= v }
        """
        return self.ancestors(v).union({v})

    def descendants(self, v):
        """
        Return the set { u : u < v }
        """
        return self.descendants_.adjacencies(v)

    def downset(self, v):
        """
        Return the set { u : u <= v }
        """
        return self.descendants(v).union({v})

    def interval(self, p, q):
        """
        Return the minimal interval that has p and q
        """
        if p < q:
            return self.downset(q).intersection(self.upset(p))
        else:
            return self.downset(p).intersection(self.upset(q))

    def isConvexSet(self, I):  # return true if I is a convex subset in poset P
        for a in I:
            for b in I:
                if a < b:
                    if not self.interval(a, b).issubset(set(I)):
                        return False
        return True

    def less(self, u, v):
        """
        Return True if u < v, False otherwise
        """
        return v in self.ancestors(u)

    def maximal(self, subset):
        """
        Return the set of elements in "subset" which are maximal
        """
        return frozenset({v for v in subset if not any(self.less(v, u) for u in subset)})

    def decreasing_list(self):
        """
        Return a list of vertices [a,b, ...] such that a > b or a not comparable with b
        """
        decresing_list_ = []
        V_ = self.vertices_
        while V_ != set():
            Max_ = self.maximal(V_)
            decresing_list_ = decresing_list_ + list(Max_)
            V_ = V_ - Max_

        return decresing_list_

    def build_viz(self, graph, shape='circle'):
        """ Return a graphviz string describing the graph and its labels """
        gv = 'digraph {\n'

        gv += f'node [fontsize=12, shape={shape}]'

        for v in graph.vertices():
            gv += f"{v}[label={v}];\n"

        for v in graph.vertices():
            for u in graph.adjacencies(v):
                gv += f"{v}->{u};\n"
        return gv + '}\n'

    def _repr_svg_(self):
        """
        Return svg representation for visual display
        """
        return graphviz.Source(self.build_viz(self.children_))._repr_svg_()


def InducedPoset(G, predicate):
    result = DirectedAcyclicGraph()
    S = set([v for v in G.vertices() if predicate(v)])
    for v in S:
        result.add_vertex(v)
    for v in S:
        for u in G.descendants(v):
            if u in S and u != v:
                result.add_edge(v, u)
    return Poset(result)



# CMGDB_util.py  # 2021-10-26
# MIT LICENSE 2020 Ewerton R. Vieira


import matplotlib
import numpy as np
import csv
import os
import CMGDB
from datetime import datetime


def base10toN(num,n):
    """N<10"""
    new_num_string=''
    current=num
    while current!=0:
        remainder=current%n
        remainder_string=str(remainder)
        new_num_string=remainder_string+new_num_string
        current=current//n
    return new_num_string





# rect = [0,0,2,2]
# subdivisions = 2
#
# print(len(FacePoints(rect, subdivisions)))
# print(FacePoints(rect, subdivisions))


class CMGDB_util:

    def __init__(self):
        self.dir_path = os.path.abspath(os.getcwd()) + "/output/"

    def Local_Lipschitz(self, X, Y):
        """Computes an estimation of the local Lipschitz constant of a given set of points"""
        dim = len(X[0])
        slope_max = [0 for i in range(dim)]
        while len(X) > 1:
            x0 = X.pop()
            x0 = np.array(x0)
            y0 = Y.pop()
            y0 = np.array(y0)
            for k, x1 in enumerate(X):
                x1 = np.array(x1)
                y1 = np.array(Y[k])
                for i in range(dim):
                    dy = np.linalg.norm(y1[i] - y0[i])
                    if np.linalg.norm(y1[i] - y0[i]) > 3:
                        continue


                    slope = dy / np.linalg.norm(x1 - x0)
                    if slope > slope_max[i]:
                        slope_max[i] = slope
        return slope_max

    def FacePoints(self, rect, subdivisions):
        """return the mid and the corners points of all faces of a rectangle"""
        dim = len(rect) // 2
        low = rect[0:dim]
        up = rect[dim::]
        box_size = [(rect[i+dim] - rect[i])/subdivisions for i in range(dim)]
        diagonal = [[rect[i] + j*box_size[i] for i in range(dim)] for j in range(1, subdivisions)]


        all = []

        for face_index in range(2**dim -1):  # all faces, remember that 2**dim is inside the cube
            face = [face_index][0]  # copy to avoid changing the indexing
            position = 0  # postion of the bits of face

            face_points = []
            while face or position < dim:
                face_points_temp = face_points.copy()   # copy to avoid changing the indexing
                face_points = []
                if not face_points_temp:
                    face_points_temp = [[]]

                if face & 1:  # if 1 than subdivision
                    for point in face_points_temp:
                        for diag_element in diagonal:
                            face_points.append(point + [diag_element[position]])

                else:  # add low or up elements of the bounds
                    for point in face_points_temp:
                        face_points.append(point + [low[position]])
                        face_points.append(point + [up[position]])

                face = face >> 1  # next bit
                position += 1  # next bit position

            all += face_points

        return all

    def FacePoints_old(self, rect):
        """return the mid and the corners points of all faces of a rectangle"""
        dim = len(rect) // 2
        face = rect + [(rect[i+dim] + rect[i])/2 for i in range(dim)]
        all = []
        for i in range(3**dim):
            ii = base10toN(i,3)
            ii = ii[::-1] #invert
            ii += '0'*(dim - len(ii)) #add 0
            all.append([face[j + int(ii[j])*dim] for j in range(dim)])
        return all

    def run_CMGDB(self, subdiv_min, subdiv_max, lower_bounds, upper_bounds, phase_periodic, F, base_name, subdiv_init=6, subdiv_limit=10000, cmap=matplotlib.cm.get_cmap('viridis', 256)):
        # Define the parameters for CMGDB

        model = CMGDB.Model(subdiv_min, subdiv_max, subdiv_init, subdiv_limit,
                            lower_bounds, upper_bounds, phase_periodic, F)

        startTime = datetime.now()

        morse_graph, map_graph = CMGDB.ComputeMorseGraph(model)

        print(datetime.now() - startTime)

        # Save Morse graph
        
        MG = self.dir_path + base_name
        morse_fname = self.dir_path + base_name + ".csv"

        CMGDB.PlotMorseGraph(morse_graph, cmap).format = 'png'
        CMGDB.PlotMorseGraph(morse_graph, cmap).render(MG)

        # Save file
        morse_nodes = range(morse_graph.num_vertices())
        morse_sets = [box + [node]
                      for node in morse_nodes for box in morse_graph.morse_set_boxes(node)]
        np.savetxt(morse_fname, np.array(morse_sets), delimiter=',')

        return morse_graph, map_graph

    # def run_CMGDB(phase_subdiv, lower_bounds, upper_bounds, phase_periodic, F, base_name):
    #     # Define the parameters for CMGDB
    #
    #     model = CMGDB.Model(phase_subdiv, lower_bounds, upper_bounds, phase_periodic, F)
    #
    #     startTime = datetime.now()
    #
    #     morse_graph, map_graph = CMGDB.ComputeMorseGraph(model)
    #
    #     print(datetime.now() - startTime)
    #
    #     # Save Morse graph
    #     self.dir_path = os.path.abspath(os.getcwd()) + "/output/"
    #     MG = self.dir_path + base_name
    #     morse_fname = self.dir_path + base_name + ".csv"
    #
    #     CMGDB.PlotMorseGraph(morse_graph).format = 'png'
    #     CMGDB.PlotMorseGraph(morse_graph).render(MG)
    #
    #     # Save file
    #     morse_nodes = range(morse_graph.num_vertices())
    #     morse_sets = [box + [node] for node in morse_nodes for box in morse_graph.morse_set_boxes(node)]
    #     np.savetxt(morse_fname, np.array(morse_sets), delimiter=',')
    #
    #     return morse_graph, map_graph

    def F_K(self, f, rect, K):
        """Input: function f, rectangle rect, and the Lipschit constant in vector form K
        Output: Image of rectangle by f, also taking in account the expansion given by K"""
        half = len(rect) // 2
        im_center = f([rect[i] + (rect[half + i] - rect[i])/2 for i in range(half)])
        list1 = []
        list2 = []
        for i in range(half):  # image of the center of the rect +or- lenght * K
            list1.append(im_center[i] - (rect[half + i] - rect[i]) * K[i] / 2)
            list2.append(im_center[i] + (rect[half + i] - rect[i]) * K[i] / 2)
        return list1 + list2

    def Morse_sets_vol(self, name_file):
        """Compute the volume of a Morse set"""
        with open(name_file, 'r') as f:
            lines = csv.reader(f, delimiter=',')
            d_vol = dict()
            for row in lines:
                size = len(row) - 1
                half = int(size / 2)
                volume_cube = 1
                for i in range(half):
                    volume_cube *= float(row[half + i]) - float(row[i])
                if row[size] in d_vol.keys():
                    d_vol[row[size]] = d_vol[row[size]] + volume_cube
                else:
                    d_vol[row[size]] = volume_cube
        return d_vol

    def sample_points(self, lower_bounds, upper_bounds, num_pts):
        # Sample num_pts in dimension dim, where each
        # component of the sampled points are in the
        # ranges given by lower_bounds and upper_bounds
        dim = len(lower_bounds)
        X = np.random.uniform(lower_bounds, upper_bounds, size=(num_pts, dim))
        return X

    def BoxMapK(self, f, rect, K):
        dim = int(len(rect) / 2)
        X = CMGDB.CornerPoints(rect)
        # Evaluate f at point in X
        Y = [f(x) for x in X]
        # Get lower and upper bounds of Y
        Y_l_bounds = [min([y[d] for y in Y]) - K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
        Y_u_bounds = [max([y[d] for y in Y]) + K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
        return Y_l_bounds + Y_u_bounds


    def BoxMapK_valid(self, f, rect, K, valid_list, point2cell):
        """Box map for valid cells based on a list using point2cell indexing"""
        id = point2cell(CMGDB.CenterPoint(rect)[0]) # center to avoid boundary points
        dim = len(rect) // 2
        if valid_list[id]:
            X = self.FacePoints(rect, 2)
            # X = CMGDB.CornerPoints(rect)
            Y = [f(x) for x in X]
            # Get lower and upper bounds of Y
            Y_l_bounds = [min([y[d] for y in Y]) - K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
            Y_u_bounds = [max([y[d] for y in Y]) + K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
            return Y_l_bounds + Y_u_bounds
        else:
            return [30000]*2*dim

    def BoxMapK_AE(self, f, rect, K, valid_state):
        """Box map for autoencoder"""
        dim = int(len(rect) / 2)
        X = CMGDB.CornerPoints(rect)
        # Evaluate f at point in X
        # print(X)
        Y = [f(x) for x in X if valid_state(x)]
        # print(Y)

        # print(Y)

        # if not Y: return [0]*2*dim
        # if len(Y)<4: return [3]*2*dim#[-.5, .25, -.5+0.01, .25+0.01]#[0]*2*dim
        if len(Y)<4 or np.linalg.norm(np.array(X[0]))<0 : return [3]*2*dim#[-.5, .25, -.5+0.01, .25+0.01]#[0]*2*dim

        # Get lower and upper bounds of Y
        Y_l_bounds = [min([y[d] for y in Y]) - K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
        Y_u_bounds = [max([y[d] for y in Y]) + K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
        return Y_l_bounds + Y_u_bounds


    def Box_GP_K(self, learned_f, rect, K, n=-3):
        """learned_f with predicted mean and standard deviation
        K Lipschit constant"""
        dim = int(len(rect) / 2)
        X = CMGDB.CornerPoints(rect)
        # print(X)
        # Evaluate f at point in X

        Y, S = learned_f(X[0])

        X = X[1::]
        for x_ in X:
            y_, s_ = learned_f(x_)
            Y = np.concatenate((Y, y_))
            S = np.concatenate((S, s_))

        # print(f"{Y}\n {S} \n {S_}")
        Y_max = Y + S*(2**n)
        Y_min = Y - S*(2**n)

        # print(f"{Y_min}\n {Y_max}")

        # print(f"{np.amin(Y_min[:,0])}\n {np.amax(Y_max[:,1])}")

        # Get lower and upper bounds of Y
        Y_l_bounds = [np.amin(Y_min[:, d]) - K*(rect[d + dim] - rect[d]) for d in range(dim)]
        Y_u_bounds = [np.amax(Y_max[:, d]) + K*(rect[d + dim] - rect[d]) for d in range(dim)]
        return Y_l_bounds + Y_u_bounds

    def F_GP_K(self, learned_f, rect, K, n=-3):
        """learned_f with predicted mean and standard deviation
        K Lipschit constant"""
        dim = int(len(rect) / 2)
        X = CMGDB.CenterPoint(rect)
        # print(X)
        # Evaluate f at point in X
        Y, S = learned_f(X)

        # print(f"{Y}\n {S} \n {S_}")
        Y_max = Y + S*(2**n)
        Y_min = Y - S*(2**n)

        # print(f"{Y_min}\n {Y_max}")

        # print(f"{np.amin(Y_min[:,0])}\n {np.amax(Y_max[:,1])}")

        # Get lower and upper bounds of Y
        Y_l_bounds = [np.amin(Y_min[:, d]) - K*(rect[d + dim] - rect[d]) for d in range(dim)]
        Y_u_bounds = [np.amax(Y_max[:, d]) + K*(rect[d + dim] - rect[d]) for d in range(dim)]
        return Y_l_bounds + Y_u_bounds

    def Box_J(self, f, J, rect):
        """f: function, J: Jacobian matrix, rect: rectangle
        Given a rect return the smallest rectangle that contains the image of the
        J \cdot rect"""
        dim = int(len(rect) / 2)
        x = rect[0:dim]
        y = f(x)
        Jac, _ = J(np.array(x).reshape(-1, dim))
        # next, add the sum of the columns
        Jac = np.concatenate((Jac, Jac.sum(axis=0).reshape(1, dim)), axis=0)

        Y_l_bounds = []
        Y_u_bounds = []
        for d in range(dim):
            Y_l_bounds.append(
                np.amin(Jac[:, d]) * (rect[d + dim] - rect[d]) + y[d]
            )

            Y_u_bounds.append(
                np.amax(Jac[:, d]) * (rect[d + dim] - rect[d]) + y[d]
            )

        return Y_l_bounds + Y_u_bounds

    def F_J(self, learned_f, J, rect, lower_bounds, n=-3, weak_multivalued_map=1):
        """f: function, J: Jacobian matrix, rect: rectangle
        Given a rect return the smallest rectangle that contains the image of the
        J \cdot rect
        weak_multivalued_map: the number to compute the subset S of all cells"""
        dim = int(len(rect) / 2)
        X = rect[0:dim]
        Y, S = learned_f(X)
        Y_max = Y + S*(2**n)
        Y_min = Y - S*(2**n)

        size_of_box = [rect[d+dim] - rect[d] for d in range(dim)]

        if weak_multivalued_map:
            coordinate = [int(
                np.rint((rect[d] - lower_bounds[d]) / size_of_box[d])
            ) % (1+weak_multivalued_map) for d in range(dim)
            ]
        else:
            coordinate = True

        # print(coordinate)
        Y_l_bounds = []
        Y_u_bounds = []

        # print(f"X {X} \n Y {Y} \n Y_min {Y_min}\n Y_max {Y_max} \n size_of_box {size_of_box}")

        if not all(coordinate):  # Compute J if all coordinate = 0 module (weak_multivalued_map+1)

            print(coordinate)
            Jac, _ = J(np.array(X).reshape(-1, dim))

            # print(f"J {Jac}")
            for d in range(dim):
                J_d_norm = np.linalg.norm(Jac[d, :])
                Y_l_bounds.append(Y_min[:, d] - J_d_norm * size_of_box[d])
                Y_u_bounds.append(Y_max[:, d] + J_d_norm * size_of_box[d])

        else:

            for d in range(dim):
                Y_l_bounds.append(Y_min[:, d])
                Y_u_bounds.append(Y_max[:, d])

        return Y_l_bounds + Y_u_bounds

    def Box_ptwise(self, learned_f, rect, n=-3):
        """learned_f with predicted mean applied to the corner points
        and standard deviation applied to the center point"""

        dim = int(len(rect) / 2)
        X = CMGDB.CenterPoint(rect) + CMGDB.CornerPoints(rect)
        # Evaluate f at point in X

        Y, S = learned_f(X[0])

        Y = Y - S*(2**n)
        Y = np.concatenate((Y, Y + 2 * S * (2 ** n)))

        X = X[1::]
        for x_ in X:
            y_, _ = learned_f(x_)
            Y = np.concatenate((Y, y_))

        # Get lower and upper bounds of Y
        Y_l_bounds = [np.amin(Y[:, d]) for d in range(dim)]
        Y_u_bounds = [np.amax(Y[:, d]) for d in range(dim)]
        return Y_l_bounds + Y_u_bounds

    def Box_noisy_K(self, f, rect, K, noise):
        """Box map with noise = [x_epsilon, y_epsilon, z_epsilon, f_epsilon]"""

        noise_x = noise[0:-1]
        noise_f = noise[-1]

        dim = int(len(rect) / 2)
        X = CMGDB.CornerPoints(rect)
        # X = self.FacePoints(rect, 2)
        # Evaluate f at point in X
        Y = [f(x) for x in X]

        # K = 2 * self.Local_Lipschitz(X, Y)

        # Get lower and upper bounds of Y
        Y_l_bounds = [min([y[d] for y in Y]) - K[d] * (rect[d + dim] - rect[d] + noise_x[d]) - noise_f for d in range(dim)]
        Y_u_bounds = [max([y[d] for y in Y]) + K[d] * (rect[d + dim] - rect[d] + noise_x[d]) + noise_f for d in range(dim)]
        f_rect = Y_l_bounds + Y_u_bounds
        return f_rect

    def Box_noisy_K_wa(self, f, rect, K, noise):
        """Box map with noise = [x_epsilon, y_epsilon, z_epsilon, f_epsilon]"""

        noise_x = noise[0:-1]
        noise_f = noise[-1]

        dim = int(len(rect) / 2)
        X = CMGDB.CornerPoints(rect)
        # X = self.FacePoints(rect, 2)
        # Evaluate f at point in X
        Y = np.array([f(x) for x in X])

        # K = 2 * self.Local_Lipschitz(X, Y)
        
        Amin = np.amin(Y, axis=0)
        Amax = np.amax(Y, axis=0)
        Y_l_bounds = [Amin[d] - K[d] * (rect[d + dim] - rect[d] + noise_x[d]) - noise_f for d in range(dim)]
        Y_u_bounds = [Amax[d] + K[d] * (rect[d + dim] - rect[d] + noise_x[d]) + noise_f for d in range(dim)]
        if Amax[0] - Amin[0]>3.14:
            # print("DIFF", Amax[0] - Amin[0])
            Y_l_bounds[0] = Y_u_bounds[0]
            Y_u_bounds[0] = 3.14158
        f_rect = Y_l_bounds + Y_u_bounds
        return f_rect

    def F_data(self, rect, id2image, point2cell, K):
        dim = len(rect) // 2
        id_of_rect = point2cell(CMGDB.CenterPoint(rect)[0]) # center to avoid boundary points
        Y = id2image[id_of_rect]
        if Y:
            # Y is a list of a array
            # Get lower and upper bounds of Y
            Amin = np.amin(Y, axis=0)
            Amax = np.amax(Y, axis=0)
            Y_l_bounds = [Amin[d] - K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
            Y_u_bounds = [Amax[d] + K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
            f_rect = Y_l_bounds + Y_u_bounds
            return f_rect
        else:
            return [30000]*2*dim

    def F_data_wa(self, rect, id2image, point2cell, K):
        dim = len(rect) // 2
        center = CMGDB.CenterPoint(rect)[0]
        # box_removed = [-0.1, 0.07, 0.1, 0.2]
        # if all([box_removed[d] < center[d] < box_removed[d+dim] for d in range(dim)]):
        #     return [30000]*2*dim
        # box_removed = [-1.5, -0.1, -0.1, 1.6]
        # if all([box_removed[d] < center[d] < box_removed[d+dim] for d in range(dim)]):
        #     return [30000]*2*dim
        # box_removed = [0.1, -0.1, 1.5, 1.6]
        # if all([box_removed[d] < center[d] < box_removed[d+dim] for d in range(dim)]):
        #     return [30000]*2*dim           
        id_of_rect = point2cell(center) # center to avoid boundary points
        Y = id2image[id_of_rect]
        

        if Y:
            # Y is a list of a array
            # Get lower and upper bounds of Y
            Amin = np.amin(Y, axis=0)
            Amax = np.amax(Y, axis=0)
            Y_l_bounds = [Amin[d] - K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
            Y_u_bounds = [Amax[d] + K[d]*(rect[d + dim] - rect[d]) for d in range(dim)]
            if Amax[0] - Amin[0]>3.14:
                # print("DIFF", Amax[0] - Amin[0])
                Y_l_bounds[0] = Y_u_bounds[0]
                Y_u_bounds[0] = 3.14159


            f_rect = Y_l_bounds + Y_u_bounds
            return f_rect
        else:
            return [30000]*2*dim


# RoA.py  # 2022-20-01
# MIT LICENSE 2020 Ewerton R. Vieira

import pychomp
import CMGDB

import numpy as np
import csv
import os

import matplotlib
import dytop.Poset as Poset
import dytop.Morse_graph_updated as Morse_graph_updated
import dytop.PlotRoA as PlotRoA


class RoA:

    def propagate_adaptive(self, u, adjacencies):
        """propagate a subtree with root u and
        assign the maximal morse node in the subtree for each tile. Inclui adaptive grid"""

        adjacencies_morse_node = set()  # save morse node assigned to each adjacent tile
        for w in adjacencies:
            morse_node = self.dict_tiles.get(w, None)  # get the morse node assigned to tile

            if morse_node == None:  # if there isnt morse node assigned then propagate
                morse_node = self.propagate_adaptive(w, self.map_graph.adjacencies(w))
            adjacencies_morse_node.add(morse_node)  # add the morse node assigned to tile w

        # get one of the maximal morse node in the subtree, it can have more than one,
        # so here we select the first ()
        # tiles that are mapped outside are assign to a fake node with value equal to -1

        adjacencies_morse_node = adjacencies_morse_node - {-1}
        max = list(self.MG.maximal(adjacencies_morse_node))

        if max:
            morse_node = max[0]

            if len(max) == 1:  # assign only for unique since we cant create more relations
                for new_morse_node in adjacencies_morse_node - self.MG_updated.adjacencies(morse_node):
                    self.MG_updated.add_edge(morse_node, new_morse_node)

        else:
            morse_node = -1

        self.dict_tiles[u] = morse_node
        if u in self.S:  # remove it since we dont have to compute again
            self.S.remove(u)
        return morse_node

    def assign_morse_nodes2tiles_adaptive(self):
        """For each tile assign a morse node (creating a region of attraction for a downset)
         Inclui adaptive grid"""

        # It will create a updated Morse graph
        self.MG_updated = pychomp.DirectedAcyclicGraph()
        for vertice in range(self.morse_graph.num_vertices()):
            self.MG_updated.add_vertex(vertice)


        # clone self.tiles_in_morse_sets we dont have to recompute
        self.dict_tiles = dict(self.tiles_in_morse_sets)

        # keep the keys in self.tiles_in_morse_sets we have to recompute
        #TODO: remove keys that correspond to attractors
        self.S = list(self.vertices())

        while self.S:  # loop: assign morse node to a tile and remove it
            v = self.S[0]
            # propagate to determine which morse node should be assign to tile v
            morse_node = self.propagate_adaptive(v, self.map_graph.adjacencies(v))


        # self.MG_updated = self.MG_updated.transitive_reduction
        self.morse_graph_updated = Morse_graph_updated.Morse_graph_updated(self.morse_graph, self.MG_updated)

        return self.dict_tiles

    def propagate(self, u, adjacencies):
        """propagate a subtree with root u and
        assign the maximal morse node in the subtree for each tile"""

        adjacencies_morse_node = set()  # save morse node assigned to each adjacent tile
        for w in adjacencies:
            morse_node = self.dict_tiles.get(w, None)  # get the morse node assigned to tile

            if morse_node == None:  # if there isnt morse node assigned then propagate
                morse_node = self.propagate(w, self.map_graph.adjacencies(w))
            adjacencies_morse_node.add(morse_node)  # add the morse node assigned to tile w

        # get one of the maximal morse node in the subtree, it can have more than one,
        # so here we select the first ()
        # tiles that are mapped outside are assign to a fake node with value equal to -1

        max = list(self.MG.maximal(adjacencies_morse_node - {-1}))
        morse_node = max[0] if max else -1

        self.dict_tiles[u] = morse_node
        if u in self.S:  # remove it since we dont have to compute again
            self.S.remove(u)
        return morse_node

    def assign_morse_nodes2tiles(self):
        """For each tile assign a morse node (creating a region of attraction for a downset)"""

        # clone self.tiles_in_morse_sets we dont have to recompute
        self.dict_tiles = dict(self.tiles_in_morse_sets)

        # remove keys in self.tiles_in_morse_sets we dont have to recompute
        self.S = list(self.vertices() - set(self.tiles_in_morse_sets.keys()))

        while self.S:  # loop: assign morse node to a tile and remove it
            v = self.S[0]
            # propagate to determine which morse node should be assign to tile v
            morse_node = self.propagate(v, self.map_graph.adjacencies(v))

        return self.dict_tiles

    def __init__(self, map_graph, morse_graph, adaptive=False):
        """
        Region of Attraction class
        Assign cells in the phase space that are mapped to a unique Morse Node
        (Regions that are uniquely mapped to Morse Sets).
        Equivalent to an order retraction onto the Morse tiles by mapping
        to unique successor.
        Input: adaptive = True: Adaptive grid and update Morse graph
        """

        self.dir_path = os.path.abspath(os.getcwd()) + "/output/"

        self.morse_graph = morse_graph
        self.map_graph = map_graph

        # Get number of vertices
        self.num_verts = map_graph.num_vertices()

        self.vertices_ = {a for a in range(self.num_verts)}

        self.tiles_in_morse_sets = {}

        # create a dict: tile in morse set -> morse node
        # it is need to create the condensation graph
        for i in range(self.morse_graph.num_vertices()):
            for j in self.morse_graph.morse_set(i):

                self.tiles_in_morse_sets[j] = i

        cyclic_Morse_graph = False
        MG = pychomp.DirectedAcyclicGraph()  # building Morse Graph Poset
        MG.add_vertex(0)
        for u in range(morse_graph.num_vertices()):
            for v in morse_graph.adjacencies(u):
                MG.add_edge(u, v)
                if u in MG.adjacencies(v):  # prevent to compute with cyclic MG
                    cyclic_Morse_graph = True

        if cyclic_Morse_graph:
            print("\033[91m Morse Graph input is cyclic, wrong input \033[0m")
            self.dict_tiles = False

        else:
            self.MG = Poset.Poset(MG)


            if adaptive:
                self.assign_morse_nodes2tiles_adaptive()
            else:
                self.assign_morse_nodes2tiles()

        # print(
        #     f"memory size of: tiles_in_morse_sets={asizeof(self.tiles_in_morse_sets)}\n",
        #     f"memory size of: dict_tiles={asizeof(self.dict_tiles)}\n",
        #     f"MG={asizeof(self.MG)}")

        # newcolors
        viridis = matplotlib.cm.get_cmap('viridis', 256)
        newcolors = viridis(np.linspace(0, 1, 256))
        orange = np.array([253/256, 174/256, 97/256, 1])
        yellowish = np.array([233/256, 204/256, 50/256, 1])
        newcolors[109:146, :] = orange
        newcolors[219:, :] = yellowish
        self.newcmp = matplotlib.colors.ListedColormap(newcolors)
        # self.newcmp=matplotlib.cm.brg  # old default option

    def vertices(self):
        """
        Return the set of elements in the poset
        """
        return self.vertices_

    def box_center(self, rect):
        dim = len(rect) // 2
        return [rect[i] + (rect[dim + i] - rect[i])/2 for i in range(dim)]

    def save_file(self, name=""):
        rect = self.morse_graph.phase_space_box(0)
        dim = int(len(rect) // 2)
        size_box = [rect[dim + i] - rect[i] for i in range(dim)]

        # getting the bounds
        lower_bounds = rect[0:dim]
        rect = self.morse_graph.phase_space_box(self.map_graph.num_vertices()-1)
        upper_bounds = rect[dim::]

        name = self.dir_path + name + "_RoA_" + ".csv"
        with open(name, "w") as file:
            f = csv.writer(file)
            f.writerow(["Box size", "Lower bounds", "Upper bounds"])
            f.writerow(size_box+lower_bounds+upper_bounds)
            f.writerow(["Tile", "Morse_node", "Box"])

            if self.dict_tiles:  # it might be empty
                # tiles in roa
                for tile_in_roa in set(self.dict_tiles.items()) - set(self.tiles_in_morse_sets.items()):
                    tile_in_roa = list(tile_in_roa) + \
                        [a for a in self.morse_graph.phase_space_box(tile_in_roa[0])]
                    f.writerow(tile_in_roa)
                # tiles in morse sets
                f.writerow(["Tile_in_Morse_set", "Morse_node", "Box"])
                for tile_in_morse_set in self.tiles_in_morse_sets.items():
                    tile_in_morse_set = list(tile_in_morse_set) + \
                        [a for a in self.morse_graph.phase_space_box(tile_in_morse_set[0])]
                    f.writerow(tile_in_morse_set)

    def Morse_sets_vol(self):
        """Compute a dict that gives the volume of the regions of attraction"""

        d_vol = dict()
        tiles_and_morse_nodes = list(self.dict_tiles.items())
        for tile_and_morse in tiles_and_morse_nodes:
            i, j = tile_and_morse  # i is the tile and j is the associated morse node
            box = self.morse_graph.phase_space_box(i)

            size = len(box)
            half = int(size / 2)

            volume_cube = 1
            for k in range(half):
                volume_cube *= float(box[half + k]) - float(box[k])

            d_vol[j] = d_vol.get(j, 0) + volume_cube

        return d_vol

    def PlotRoA(self, selection=[], fig_w=8, fig_h=8, xlim=None, ylim=None,
                  cmap=matplotlib.cm.get_cmap('viridis', 256), name_plot='', from_file=None, plot_point=False, section=None):

        if not os.path.exists(self.dir_path):
            os.makedirs(self.dir_path)
        self.save_file(name="temp")

        rect = self.morse_graph.phase_space_box(0)
        dim = int(len(rect) // 2)

        # getting the bounds
        lower_bounds = rect[0:dim]
        upper_bounds = rect[dim::]

        for i in range(self.num_verts):
            box = self.morse_graph.phase_space_box(i)
            for index, j in enumerate(box[0:dim]):
                if lower_bounds[index] > j:
                    lower_bounds[index] = j
            for index, j in enumerate(box[dim::]):
                if upper_bounds[index] < j:
                    upper_bounds[index] = j

        fig, ax = PlotRoA.PlotRoA(lower_bounds, upper_bounds, selection=selection, fig_w=fig_w, fig_h=fig_h, xlim=xlim,
                            ylim=ylim, cmap=cmap, name_plot=name_plot, from_file="temp", plot_point=plot_point, section=section)

        os.remove(self.dir_path + "temp_RoA_.csv")
        return fig, ax




from setuptools import setup

setup (
    name = "dytop",
    version = "0.1.9",
    author = "Ewerton Rocha Vieira",
    url = "https://github.com/Ewerton-Vieira/dytop.git",
    description = "dytop: combinatorial DYnamics and TOPology",
    long_description = open('README.md').read(),
    long_description_content_type='text/markdown',
    ext_package='dytop',
    packages=['dytop'],
    install_requires = ['numpy', 'scipy', 'matplotlib', 'CMGDB', 'pychomp2', 'matplotlib']
)